#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Security.Policy;
using System.Reflection;
using System.Diagnostics;
using System.Security.Permissions;
using System.Security;
using System.Linq;
using HeuristicLab.PluginInfrastructure.Manager;
using System.IO;

namespace HeuristicLab.PluginInfrastructure {

  /// <summary>
  /// Lightweight application manager is set as the application manager as long as the plugin infrastructure is uninitialized.
  /// The list of plugins and applications is empty. The default application manager is necessary to provide the type discovery
  /// functionality in unit tests.
  /// </summary>
  internal sealed class LightweightApplicationManager : IApplicationManager {

    #region IApplicationManager Members
    /// <summary>
    /// Gets an empty list of plugins. (LightweightApplicationManager doesn't support plugin discovery)
    /// </summary>
    public IEnumerable<IPluginDescription> Plugins {
      get { return new IPluginDescription[0]; }
    }

    /// <summary>
    /// Gets an empty list of applications. (LightweightApplicationManager doesn't support application discovery)
    /// </summary>
    public IEnumerable<IApplicationDescription> Applications {
      get { return new IApplicationDescription[0]; }
    }

    /// <summary>
    /// Creates an instance of all types that are subtypes or the same type of the specified type
    /// </summary>
    /// <typeparam name="T">Most general type.</typeparam>
    /// <returns>Enumerable of the created instances.</returns>
    public IEnumerable<T> GetInstances<T>() where T : class {
      return GetInstances(typeof(T)).Cast<T>();
    }

    /// <summary>
    /// Creates an instance of all types that are subtypes or the same type of the specified type
    /// </summary>
    /// <param name="type">Most general type.</param>
    /// <returns>Enumerable of the created instances.</returns>
    public IEnumerable<object> GetInstances(Type type) {
      return from t in GetTypes(type, true)
             select Activator.CreateInstance(t);
    }


    /// <summary>
    /// Finds all instantiable types that are subtypes or equal to the specified type.
    /// </summary>
    /// <param name="type">Most general type for which to find matching types.</param>
    /// <remarks>Return only types that are instantiable 
    /// (interfaces, abstract classes... are not returned)</remarks>
    /// <returns>Enumerable of the discovered types.</returns>
    public IEnumerable<Type> GetTypes(Type type) {
      return GetTypes(type, true);
    }

    /// <summary>
    /// Finds all types that are subtypes or equal to the specified type.
    /// </summary>
    /// <param name="type">Most general type for which to find matching types.</param>
    /// <param name="onlyInstantiable">Return only types that are instantiable 
    /// (interfaces, abstract classes... are not returned)</param>
    /// <returns>Enumerable of the discovered types.</returns>
    public IEnumerable<Type> GetTypes(Type type, bool onlyInstantiable) {
      return from asm in AppDomain.CurrentDomain.GetAssemblies()
             from t in GetTypes(type, asm, onlyInstantiable)
             select t;
    }

    /// <summary>
    /// Gets types that are assignable (same of subtype) to the specified type only from the given assembly.
    /// </summary>
    /// <param name="type">Most general type we want to find.</param>
    /// <param name="assembly">Assembly that should be searched for types.</param>
    /// <param name="onlyInstantiable">Return only types that are instantiable 
    /// (interfaces, abstract classes...  are not returned)</param>
    /// <returns>Enumerable of the discovered types.</returns>
    private static IEnumerable<Type> GetTypes(Type type, Assembly assembly, bool onlyInstantiable) {
      return from t in assembly.GetTypes()
             where type.IsAssignableFrom(t)
             where onlyInstantiable == false || (!t.IsAbstract && !t.IsInterface && !t.HasElementType)
             select t;
    }

    /// <summary>
    /// Not supported by the LightweightApplicationManager
    /// </summary>
    /// <param name="type"></param>
    /// <param name="plugin"></param>
    /// <returns></returns>
    /// <throws>NotSupportedException</throws>
    public IEnumerable<Type> GetTypes(Type type, IPluginDescription plugin) {
      throw new NotSupportedException("LightweightApplicationManager doesn't support type discovery for plugins.");
    }

    /// <summary>
    /// Not supported by the LightweightApplicationManager
    /// </summary>
    /// <param name="type"></param>
    /// <param name="plugin"></param>
    /// <param name="onlyInstantiable"></param>
    /// <returns></returns>
    /// <throws>NotSupportedException</throws>
    public IEnumerable<Type> GetTypes(Type type, IPluginDescription plugin, bool onlyInstantiable) {
      throw new NotSupportedException("LightweightApplicationManager doesn't support type discovery for plugins.");
    }

    /// <summary>
    /// Not supported by the LightweightApplicationManager
    /// </summary>
    /// <param name="type"></param>
    /// <returns></returns>
    /// <throws>NotSupportedException</throws>
    public IPluginDescription GetDeclaringPlugin(Type type) {
      throw new NotSupportedException("LightweightApplicationManager doesn't support type discovery for plugins.");
    }

    #endregion
  }
}
